@@ -1,4 +1,4 @@
-# Copyright (c) 2022, Riverbank Computing Limited
+# Copyright (c) 2023, Riverbank Computing Limited
 # All rights reserved.
 #
 # This copy of SIP is licensed for use under the terms of the SIP License
@@ -21,15 +21,15 @@
 # POSSIBILITY OF SUCH DAMAGE.
 
 
+from ..scoped_name import ScopedName
 from ..specification import (AccessSpecifier, Argument, ArgumentType,
         ArrayArgument, ClassKey, Docstring, DocstringFormat, Extract,
         FunctionCall, IfaceFile, IfaceFileType, KwArgs, License, MappedType,
-        MappedTypeTemplate, Module, Overload, Property, PyQtMethodSpecifier,
-        QualifierType, ScopedName, Signature, Template, ThrowArguments, Value,
-        ValueType, VirtualErrorHandler, WrappedClass, WrappedTypedef,
-        WrappedVariable)
+        MappedTypeTemplate, Overload, Property, PyQtMethodSpecifier,
+        QualifierType, Signature, Template, ThrowArguments, Value, ValueType,
+        VirtualErrorHandler, WrappedTypedef, WrappedVariable)
 from ..templates import same_template_signature
-from ..utils import normalised_scoped_name, search_typedefs
+from ..utils import cached_name, normalised_scoped_name, search_typedefs
 
 from .annotations import DottedName
 from .tokens import tokens
@@ -70,7 +70,7 @@ def p_specification(p):
 
 
 def p_statement(p):
-    """statement : EOF
+    """statement : eof
         | namespace_statement
         | composite_module
         | copying
@@ -120,6 +120,12 @@ def p_namespace_statement(p):
         | type_header_code"""
 
 
+def p_eof(p):
+    "eof : EOF"
+
+    p.parser.pm.pop_module_state()
+
+
 # State changing productions. #################################################
 
 def p_begin_args(p):
@@ -195,6 +201,12 @@ def p_composite_module(p):
     if pm.skipping:
         return
 
+    # A composite module must be the first one in the specification.
+    if not pm.in_main_module:
+        pm.parser_error(p, 1, "a %CompositeModule cannot be %Imported")
+
+    pm.spec.is_composite = True
+
     if len(p) == 4:
         name = p[2]
         body = p[3]
@@ -204,15 +216,7 @@ def p_composite_module(p):
 
     module = pm.module_state.module
 
-    if module is not pm.spec.modules[0]:
-        pm.parser_error(p, 1, "a %CompositeModule cannot be %Imported")
-
-    if module.fq_py_name is not None:
-        pm.parser_error(p, 1,
-                "%CompositeModule must appear before any %Module directive")
-
-    module.fq_py_name = pm.cached_name(str(name))
-    module.is_composite = True
+    module.fq_py_name = cached_name(pm.spec, str(name))
 
     for directive in body:
         if isinstance(directive, Docstring):
@@ -397,7 +401,7 @@ def p_defmetatype(p):
 
     symbol = 2 if len(p) == 3 else 6
 
-    module.default_metatype = pm.cached_name(str(p[symbol]))
+    module.default_metatype = cached_name(pm.spec, str(p[symbol]))
 
 
 # %DefaultSupertype ###########################################################
@@ -419,7 +423,7 @@ def p_defsupertype(p):
 
     symbol = 2 if len(p) == 3 else 6
 
-    module.default_supertype = pm.cached_name(str(p[symbol]))
+    module.default_supertype = cached_name(pm.spec, str(p[symbol]))
 
 
 # %Docstring ##################################################################
@@ -543,7 +547,7 @@ def p_extract(p):
         order = args.get('order', -1)
         part = p[7]
 
-    pm.module_state.module.extracts.append(Extract(id, order, part))
+    pm.spec.extracts.append(Extract(id, order, part.text))
 
 
 def p_extract_args(p):
@@ -879,6 +883,9 @@ def p_license_arg(p):
 # The mapped type annotations.
 _MAPPED_TYPE_ANNOTATIONS = (
     'AllowNone',
+    'NoAssignmentOperator',
+    'NoCopyCtor',
+    'NoDefaultCtor',
     'NoRelease',
     'PyName',
     'PyQtFlags',
@@ -912,6 +919,8 @@ def p_mapped_type_template(p):
 
     pm.validate_mapped_type(p, 1, pm.scope)
 
+    pm.parsing_template = False
+
     pm.pop_scope()
 
 
@@ -1024,22 +1033,15 @@ def p_module(p):
     if pm.skipping:
         return
 
-    module_state = pm.module_state
-    module = module_state.module
-
     # See if this %Module is part of a %CompositeModule.
-    if module.is_composite or module.composite is not None:
+    if pm.spec.is_composite:
         # Historically we %Include modules although conceptually we actually
         # %Import them.  Ensure that the scopes etc. are correct in either
         # case.
         pm.ensure_import()
 
-        # The module state may have changed.
-        module_state = pm.module_state
-
-        module_state.module.composite = module if module.is_composite else module.composite
-
-        module = module_state.module
+    module_state = pm.module_state
+    module = module_state.module
 
     if module.fq_py_name is not None:
         pm.parser_error(p, 1, "%Module has already been specified")
@@ -1053,7 +1055,7 @@ def p_module(p):
         body = p[7]
 
     try:
-        module.fq_py_name = pm.cached_name(str(args['name']))
+        module.fq_py_name = cached_name(pm.spec, str(args['name']))
 
         if pm.in_main_module:
             module.fq_py_name.used = True
@@ -1134,7 +1136,7 @@ def p_module_body(p):
     """module_body : '{' module_body_directives '}' ';'
         | empty"""
 
-    p[0] = p[2] if len(p) == 4 else []
+    p[0] = p[2] if len(p) == 5 else []
 
 
 def p_module_body_directives(p):
@@ -1142,12 +1144,16 @@ def p_module_body_directives(p):
         | module_body_directives module_body_directive"""
 
     if len(p) == 2:
-        value = [p[1]]
-    else:
+        body = []
         value = p[1]
-        value.append(p[2])
+    else:
+        body = p[1]
+        value = p[2]
 
-    p[0] = value
+    if value is not None:
+        body.append(value)
+
+    p[0] = body
 
 
 def p_module_body_directive(p):
@@ -1222,6 +1228,8 @@ def p_plugin(p):
     if pm.skipping:
         return
 
+    pm.deprecated(p, 1)
+
     pm.spec.plugins.append(p[2])
 
 
@@ -1254,34 +1262,34 @@ def p_preinit_code(p):
 # %Property ###################################################################
 
 def p_property(p):
-    "property : Property '(' property_args ')' '{' property_body '}' ';'"
+    "property : Property begin_args '(' property_args end_args ')' opt_property_body"
 
     pm = p.parser.pm
 
     if pm.skipping:
         return
 
-    name = p[3].get('name')
+    name = p[4].get('name')
     if name is None:
         p.parser.pm.parser_error(p, 1,
                 "a name must be specified for %Property")
         return
 
-    pm.check_attributes(p, 3, name)
+    pm.check_attributes(p, 4, name)
 
-    name = pm.cached_name(name)
+    name = cached_name(pm.spec, name)
     if pm.in_main_module:
         name.used = True
 
-    getter = p[3].get('get')
+    getter = p[4].get('get')
     if getter is None:
         p.parser.pm.parser_error(p, 1,
                 "a getter must be specified for %Property")
         return
 
-    prop = Property(name=name, getter=getter, setter=p[3].get('set'))
+    prop = Property(name=name, getter=getter, setter=p[4].get('set'))
 
-    for directive in body:
+    for directive in p[7]:
         if isinstance(directive, Docstring):
             prop.docstring = directive
 
@@ -1306,6 +1314,13 @@ def p_property_arg(p):
     p[0] = {p[1]: p[3]}
 
 
+def p_opt_property_body(p):
+    """opt_property_body : empty
+        | '{' property_body '}' ';'"""
+
+    p[0] = [] if len(p) == 2 else p[2]
+
+
 def p_property_body(p):
     """property_body : property_line
         | property_body property_line"""
@@ -1517,6 +1532,12 @@ def p_cpp_type(p):
         value.derefs.extend(p[2])
         value.is_reference = p[3]
 
+        # PyObject * is a synonym for SIP_PYOBJECT.
+        if value.type is ArgumentType.DEFINED and len(value.definition) == 1 and value.definition.base_name == 'PyObject' and len(value.derefs) == 1 and not value.is_reference:
+            value.type = ArgumentType.PYOBJECT
+            value.definition = None
+            value.derefs = []
+
     p[0] = value
 
 
@@ -1527,11 +1548,11 @@ def p_base_type(p):
         | struct scoped_name
         | union scoped_name"""
 
+    pm = p.parser.pm
+
     if isinstance(p[1], ArgumentType):
         p[0] = Argument(p[1])
     elif len(p) == 2:
-        pm = p.parser.pm
-
         # Resolve it if it is the name of a typedef.  This is done early as a
         # workaround for allowing /PyInt/ to be applied to typedef'ed types.
         ad = Argument(ArgumentType.DEFINED, definition=p[1])
@@ -1561,6 +1582,8 @@ def p_base_type(p):
         p[0] = Argument(ArgumentType.TEMPLATE,
                 definition=Template(p[1], Signature(args=p[3])))
 
+    p[0].source_location = pm.get_source_location(p, 1)
+
 
 # Map unsigned two-word POD types.
 _UNSIGNED_MAP = {
@@ -1642,6 +1665,9 @@ def p_pod_type(p):
         else:
             p[0] = _UNSIGNED_MAP[p[2]]
     else:
+        if p[1] == 'SIP_SSIZE_T':
+            p.parser.pm.deprecated(p, 1, instead="'Py_ssize_t'")
+
         p[0] = _ONE_WORD_MAP[p[1]]
 
 
@@ -1725,9 +1751,26 @@ def p_class_template(p):
 
     pm.class_templates.append((p[1], p[2]))
 
+    pm.parsing_template = False
+
+
+def p_class_docstring(p):
+    "class_docstring : docstring"
+
+    pm = p.parser.pm
+
+    if pm.skipping:
+        return
+
+    if pm.scope.docstring is None:
+        pm.scope.docstring = p[1]
+    else:
+        pm.parser_error(p, 1,
+                "%Docstring has already been defined for this class")
+
 
 def p_class_decl(p):
-    "class_decl : class class_head opt_class_body ';'"
+    "class_decl : class class_head opt_class_definition ';'"
 
     pm = p.parser.pm
 
@@ -1758,7 +1801,7 @@ def p_class_head(p):
 
 
 def p_struct_decl(p):
-    "struct_decl : struct struct_head opt_class_body ';'"
+    "struct_decl : struct struct_head opt_class_definition ';'"
 
     pm = p.parser.pm
 
@@ -1833,7 +1876,8 @@ def p_superclass(p):
         pm.parser_error(p, 2, "super-class list contains an invalid type")
 
     # Find the actual class.
-    p[0] = pm.find_class(p, 2, IfaceFileType.CLASS, ad.definition)
+    p[0] = pm.find_class(p, 2, IfaceFileType.CLASS, ad.definition,
+            tmpl_arg=pm.parsing_template)
 
 
 def p_class_access(p):
@@ -1848,8 +1892,8 @@ def p_class_access(p):
     p[0] = p[1]
 
 
-def p_opt_class_body(p):
-    """opt_class_body : '{' opt_docstring class_body '}'
+def p_opt_class_definition(p):
+    """opt_class_definition : '{' opt_class_body '}'
         | empty"""
 
     pm = p.parser.pm
@@ -1857,11 +1901,12 @@ def p_opt_class_body(p):
     if pm.skipping:
         return
 
-    if len(p) == 5:
-        pm.scope.docstring = p[2]
-        p[0] = True
-    else:
-        p[0] = False
+    p[0] = (len(p) == 4)
+
+
+def p_opt_class_body(p):
+    """opt_class_body : class_body
+        | empty"""
 
 
 def p_class_body(p):
@@ -1873,6 +1918,8 @@ def p_class_line(p):
     """class_line : if_start
         | if_end
         | class_decl
+        | class_docstring
+        | class_template
         | ctor
         | dtor
         | enum_decl
@@ -1900,10 +1947,21 @@ def p_class_line(p):
         | type_code
         | type_header_code
         | type_hint_code
-        | BIGetReadBufferCode CODE_BLOCK
-        | BIGetWriteBufferCode CODE_BLOCK
-        | BIGetSegCountCode CODE_BLOCK
-        | BIGetCharBufferCode CODE_BLOCK"""
+        | deprecated_code_directives CODE_BLOCK"""
+
+
+def p_deprecated_code_directives(p):
+    """deprecated_code_directives : BIGetReadBufferCode
+        | BIGetWriteBufferCode
+        | BIGetSegCountCode
+        | BIGetCharBufferCode"""
+
+    pm = p.parser.pm
+
+    if pm.skipping:
+        return
+
+    pm.deprecated(p, 1)
 
 
 # The ctor annotations.
@@ -2166,7 +2224,7 @@ def p_enum_line(p):
     if len(p) == 5:
         pm.check_annotations(p, 3, "enum member", _ENUM_MEMBER_ANNOTATIONS)
 
-        p[0] = (p[1], pm.cached_name(pm.get_py_name(p[1], p[3])),
+        p[0] = (p[1], cached_name(pm.spec, pm.get_py_name(p[1], p[3])),
                 p[3].get('NoTypeHint', False))
     else:
         p[0] = None
@@ -2203,7 +2261,7 @@ def p_exception(p):
     pm.check_annotations(p, 4, "exception", _EXCEPTION_ANNOTATIONS)
 
     cpp_name = p[2]
-    py_name = pm.get_py_name(cpp_name, p[4])
+    py_name = pm.get_py_name(cpp_name.base_name, p[4])
     pm.check_attributes(p, 2, py_name)
 
     builtin_base, defined_base = p[3]
@@ -2248,9 +2306,10 @@ def p_opt_base_exception(p):
 
     if len(p) == 4:
         base = p[2]
+        base.make_absolute()
 
         # See if it is a project-defined exception.
-        for xd in p.parser_manager.spec.exceptions:
+        for xd in p.parser.pm.spec.exceptions:
             if xd.iface_file.fq_cpp_name == base:
                 defined_base = xd
                 break
@@ -2540,7 +2599,7 @@ def p_arg_type(p):
     annotations = p[3]
 
     if p[2] is not None:
-        arg.name = pm.cached_name(p[2])
+        arg.name = cached_name(pm.spec, p[2])
 
     pm.apply_common_argument_annotations(p, 3, arg, annotations)
 
@@ -2738,6 +2797,7 @@ def p_opt_exceptions(p):
 
     if p[1] == 'throw':
         p[0] = p[3]
+        p.parser.pm.deprecated(p, 1)
     elif p[1] == 'noexcept':
         p[0] = ThrowArguments()
     else:
@@ -2748,7 +2808,7 @@ def p_opt_exception_list(p):
     """opt_exception_list : exception_list
         | empty"""
 
-    p[0] = ThrowArguments(arguments=[]) if p[1] is None else p[1]
+    p[0] = ThrowArguments() if p[1] is None else p[1]
 
 
 def p_exception_list(p):
@@ -2886,7 +2946,7 @@ def p_virtual_catcher_code(p):
 
 # C++ namespaces. #############################################################
 
-# The typedef annotations.
+# The namespace annotations.
 _NAMESPACE_ANNOTATIONS = (
     'PyQtNoQMetaObject',
 )
@@ -3031,7 +3091,7 @@ _UNION_ANNOTATIONS = (
 
 
 def p_union_decl(p):
-    "union_decl : union union_head opt_class_body ';'"
+    "union_decl : union union_head opt_class_definition ';'"
 
     pm = p.parser.pm
 
@@ -3084,12 +3144,21 @@ def p_variable(p):
     annos_symbol = 3
     body = p[4]
 
+    # We don't currently support /AllowNone/ and /DisallowNone/ for variables.
+    # Generated variable setters support allow None for some (pointer) types
+    # but not others.  For pointer types we explicity disallow None.  This
+    # doesn't affect the generated code but does mean that type hints match
+    # those from previous versions.  This should be changed to properly support
+    # the annotations.
+    if len(type.derefs) != 0:
+        type.disallow_none = True
+
     annotations = p[annos_symbol]
 
     pm.check_annotations(p, annos_symbol, "variable", _VARIABLE_ANNOTATIONS)
     pm.apply_type_annotations(p, annos_symbol, type, annotations)
 
-    py_name = pm.cached_name(pm.get_py_name(cpp_name, p[3]))
+    py_name = cached_name(pm.spec, pm.get_py_name(cpp_name, p[3]))
 
     if pm.in_main_module:
         py_name.used = True
@@ -3223,6 +3292,8 @@ def p_template_decl(p):
 
     p[0] = Signature(args=p[3])
 
+    pm.parsing_template = True
+
 
 def p_bool_value(p):
     """bool_value : true
@@ -3235,18 +3306,17 @@ def p_bool_value(p):
 
 def p_dotted_name(p):
     """dotted_name : NAME
-        | dotted_name '.' NAME"""
+        | DOTTED_NAME"""
 
-    if len(p) == 2:
-        value = DottedName()
-        symbol = 1
-    else:
-        value = p[1]
-        symbol = 3
+    p[0] = DottedName(p[1])
 
-    value.append(p[symbol])
 
-    p[0] = value
+def p_file_path(p):
+    """file_path : NAME
+        | DOTTED_NAME
+        | FILE_PATH"""
+
+    p[0] = p[1]
 
 
 def p_empty(p):
@@ -3255,20 +3325,6 @@ def p_empty(p):
     p[0] = None
 
 
-def p_file_name(p):
-    """file_name : dotted_name
-        | file_name '-' dotted_name"""
-
-    p[0] = str(p[1]) if len(p) == 2 else p[1] + '-' + str(p[3])
-
-
-def p_file_path(p):
-    """file_path : file_name
-        | file_path '/' file_name"""
-
-    p[0] = p[1] if len(p) == 2 else p[1] + '/' + p[3]
-
-
 def p_opt_const(p):
     """opt_const : const
         | empty"""
